	title autotrack driver - KCT El and Az to a parallel port, resident part
;
;	Courtney Duncan, N5BF, 13 Dec 88, KCT version 09 Jan 90
;		GT support added 11 Jan 90
;
;               Timing relaxed for faster computers 15 Oct 2002 KB5MU
;
;	AutoTrack is adapted from the GrafTraks AT TSR version and 
;	modified extensively for the KCT calling format.  Command line
;	sets, calibrations, limits and inversion are supported from STOW.
;
;	Command line format:
;
;	at prn#
;			prn# is DOS lpt number for port used
;
;	Hooks into the KCT driving interrupt (63h) and strobes data
;	out for elevation and azimuth control according to tracking inputs
;	Azimuth and elevation are converted and the comparator voltages 
;	are as follows:
;
;	Az	INT((A/3)+60), A >=  0 and < 180, range 1.8 to 3.6 volts.
;		INT((A/3)-60), A > 180 and <=360, range 0.0 to 1.8 volts.
;	El	INT((E/1.5)+128), range 0 - 180 degrees, 0.0 - 3.6 volts.
;		128 is 80h, the "elevation" flag in print output.
;
;	The precalibrated limiting values can be altered.
;
;	The print driver does not read the busy lines on the interface,
;	it just writes, the rest is up to the hardware.  This prevents
;	hangups such as often occur with BASIC and other oppressive
;	software.
;
;	No Time/Target features are implemented.
;
;	The following driver calls and features are implemented:
;
;	Signature:  The interrupt vector 63h points to the execution
;	routine.  Following the entry point by 0Ah is the string "RotorDRV"
;	which is used to identify existence of a working driver.
;
;	Status:
;	Entry:  AH=0
;	Exit:   AL=0  meaning, response to I/O poll is correct,
;	              controller is enabled, nothing is stalled,
;		      nothing is disabled, nothing is moving.
;		BH=1  az/el installation
;		BL=0  computer runs in UTC
;		CH=0, current elevation not returned
;		CL=0, current azimuth not returned
;
;	Read position
;	Entry:  AH=3
;	Exit:   AX  current elevation, degrees (-90 to 269, signed)
;		BX  current azimuth, degrees (0 to 359, signed)
;		CX  target elevation, same as AX (no feedback)
;		DX  target azimuth, same as BX (no feedback)
;
;	Set position
;	Entry:  AH=4
;	Exit:   CX  target elevation (-90 to 269 degrees, signed binary)
;		DX  target azimuth (0 to 359 degrees, signed binary)
;		Both values must be in range.
;
;	All other features are ignored or returned so as to be consistent
;	with inaction as follows:
;
;	Read slew setting
;	Entry:  AH=1
;	Exit:   CH=0, no elevation setting
;		CL=0, no azimuth setting
;
;	Write slew setting
;	Entry:  AH=2, command non-destructively ignored
;	Exit:   none
;
;	Clear Time/Target table
;	Entry:  AH=5
;	Exit:   none
;
;	Get Time/Target table data
;	Entry:  AH=6
;	Exit:   AL=0, normal mode
;		CX=0, no table records
;		DX=0, no relative entry
;
;	Add Time/Target table entry
;	Entry:  AH=7
;	Exit:   AL=2, [non-existent] table full, entry not added
;
;	Read driver's date and time
;	Entry:  AH=8
;	Exit:   AX=BX=CX=DX=0, all zeroes, obviously invalid
;
;	Disable/Enable driver
;	Entry:  AH=9
;		AL ignored
;	Exit:   none
;
;	All these and other unused functions clear all registers and return.
;
;	Commands added for 'stow' user interface.
;
;	Change to invert mode, antennas "over on back"
;	Entry:	AH=96
;	Exit:   none
;
;	Set mechanical errors
;	Entry:  AH=97
;		CX= elevation error, degrees, signed binary (-9 to 9)
;		DX= azimuth error, same
;	Exit:   none
;
;	Set elevation stall limits (before mechanical error removal)
;	Entry:  AH=98
;		CX= elevation upper limit, degrees, signed binary
;		DX= elevation lower limit
;
;	Set azimuth stall limits (before mechanical error removal)
;	Entry:  AH=99
;		CX= azimuth upper limit, degrees, signed binary
;		DX= azimuth lower limit
;
;	GT support:  two functions
;
;	Write value to rotors
;	Entry:  AL=0
;		BX=azimuth
;		DX=elevation
;		CL=satellite number (1-16,90,91-92) not used
;		CH=user number (1-16) verified to be 0 (manual) or 1
;		ES:SI:DI=frequency info, not used
;	Exit:   same as Read value below
;
;	Read value of rotors (faked to previously commanded values as in KCT)
;	Entry:	AL=2
;	Exit:   BX=azimuth
;		DX=elevation

int_at	equ	63h	;interrupt service routine
int_gt	equ	104d
id_gt	equ	103d

invflp	equ	96d	;user number for inverted mode toggle
mechst	equ	97d	;user number for mechanical errors set
stlste	equ	98d	;user number for stall settings, elevation
stlsta	equ	99d	;user number for stall settings, azimuth

ATNULL	equ	999d	;value used to indicate 'no data' in KCT reads
GTNULL	equ	0FFFFh	;value used to indicate 'no data' in gt reads
;
;	macros
;

conout	macro message			;console output
	mov	dx,offset message	;message string, '$' terminator
	mov	ah,09h			;CP/M like system call
	int	21h			;invoke DOS
	endm

command	macro number,process		;command dispatch macro
	local	next
	cmp	ah,number		;determine command number
	jnz	next
	call	process			;when number found
	jmp	at_ret			;then, outta here
next:
	endm

strobe	macro				;strobe the byte in al out to port
	local	wait, wait2
	push	dx
	mov	dx,prnport		;by Murphy, dx is used for ports
	out	dx,al			;write the actual data
	inc	dx
	inc	dx			;strobe is on port C of parallel adptr.
	push	ax			;save the output value
	mov	al,0Dh			;this is how the BIOS listing does it
	out	dx,al			;strobe (bit 0) up (inverts down hdw.)
	push	cx
	mov	cx,10000d		;wait a while
wait:	loop	wait
	pop	cx
	mov	al,0Ch			;turn strobe off
	out	dx,al
       	
        push	cx
	mov	cx,10000d		;wait a while
wait2:	loop	wait2
	pop	cx

	pop	ax			;restore written value
	pop	dx
	endm

;
;	now, the installing at.com program, to be run before GT on boot
;

code	segment	public

	org	100h			;who knows why they put us 
	assume	cs:code,ds:code,es:code	; through this

;
;	load time entry point
;

start:
	push	cs
	pop	ds			;so the assume statement works

	mov	at_i,int_at		;software interrupt vectors
	mov	gt_i,int_gt
	mov	gt_id,id_gt

	conout	start_msg		;display signon message

;
;	get stuff from command line
;	ds:81h - space
;	ds:82h - lpt number, must be '1,' '2,' or '3'
;	ds:83h - CR
;
	mov	bx,81h			;1st byte should be a space
	mov	al,[bx]
	cmp	al,' '
	jne	cline_bad

	inc	bx
	mov	al,[bx]			;2nd byte
	sub	al,'0'			;valid check, must be 
	jl	cline_bad		; between 1 and 3 inclusive
	cmp	al,3d			; (also convert to binary)
	jg	cline_bad
	mov	prnnr,al		;stored

;
;	Now determine the address of the port, the last thing that can
;	go wrong. . .
;

prndet:	mov	bl,prnnr
	dec	bl			;change from 1 based to 0 based index
	xor	bh,bh			;bx now contains LPT number
	shl	bx,1			; * 2, word addressing
	push	ds
	assume	ds:nothing
	xor	ax,ax
	mov	ds,ax			;base segment, where DOS stuff is
	mov	dx,[bx][0408h]		;where DOS keeps the port numbers
	pop	ds
	assume	ds:code
	or	dx,dx			;check for zero, means no port
	jz	port_bad
	mov	prnport,dx		;save the address
	jmp	setup			;continue

;
;	invalid command line processing
;

port_bad:
	conout	badprn
	jmp	short quit

cline_bad:
	conout	usage
quit:	conout	noload
	int	20h			;return to dos, old style

;
;	the general data declaration area
;

at_i	db	?			;value of interrupt number
gt_i	db	?			;value of interrupt number
gt_id	db	?			;value of interrupt number
prnnr	db	?			;number of printer port used
onback	db	00h			;"on back" track, 00h = no, FFh = yes
prnport	dw	?			;port number for parallel output
oldaz	dw	ATNULL			;last azimuth command or none (0FFFFh)
oldel	dw	ATNULL			;last elevation command or none (0FFFFh)
elcorr	dw	0d			;mechanical elevation correction
azcorr	dw	0d			;mechanical azimuth correction
ellostl dw	0d			;minimum uncorrected value for el
elhistl dw	180d			;maximum uncorrected value for el
azlostl dw	0d			;minimum uncorrected value for az -180
azhistl dw	359d			;maximum uncorrected value for az +180
azcmd	dw	-1d			;previous word strobed
elcmd	dw	-1d			;previous word strobed

;
;	stuff to print
;

badprn	db	0Dh,0Ah,'no such printer port installed','$'
noload	db	0Dh,0Ah,0Ah,'at aborted, not loaded',0Dh,0Ah,'$'
usage	db	0Dh,0Ah,'command line must be either:'
	db	0Dh,0Ah,' '
	db	0Dh,0Ah,'   at n'
	db	0Dh,0Ah,' '
	db	0Dh,0Ah,'       where'
	db	0Dh,0Ah,'         n = printer port, 1, 2, or 3, and'
	db	0Dh,0Ah,'$'

setup:

;
;	set up the interrupt vectors and indicator
;
	mov	ah,25h			;DOS set interrupt command
	mov	al,at_i			;the interrupt service itself
	mov	dx,offset at_svc	;interrupt service routine, ds: ok
	int	21h			;do the deed

	mov	ah,25h			;DOS set interrupt command
	mov	al,gt_i			;the interrupt service itself
	mov	dx,offset gt_svc	;interrupt service routine, ds: ok
	int	21h			;do the deed

	mov	ah,25h			;DOS set interrupt command
	mov	al,gt_id		;the interrupt service itself
	mov	dx,0DEAFh		;interrupt present code, ds: ok
	int	21h			;do the deed

;
;	done, now exit, staying resident in the old style
;

tsr:
	conout	loadok			;final message
	lea	dx,at_end		;end of memory now in use
	int	27h			;tsr exit

;
;
;

start_msg:
	db	0Dh,0Ah,'IT, QT, GT to LPTn AutoTracker control loading'
	db	0Dh,0Ah,'Emulates KCT and GT int with KB5IPM hardware interface'
	db	0Dh,0Ah,' version 1.20, 11 January 90, N5BF'
	db	0Dh,0Ah,'$'

loadok:	db	0Dh,0Ah,'AutoTrack control succesfully loaded'
	db	0Dh,0Ah,'$'

;
;	here follows the actual interrupt service routine
;

at_svc	proc	far			;must be 'far' since interrupt svc.

	push	ds

	push	cs
	pop	ds			;local data, as 'assumed'
					;not using much stack, no switch

	jmp	cd
	db	'    RotorDRV'		;official signature indicates presence

cd:					;real code begins
;
;	command dispatcher
;
	command	0,at_stat		;status function
	command 1,at_null		;read slew
	command 2,at_null		;write slew
	command 3,at_read		;read antenna position commanded
	command 4,at_writ		;set antenna position
	command 5,at_null		;clear table
	command 6,at_null		;get table
	command 7,at_atbl		;add to table
	command 8,at_null		;read time
	command 9,at_null		;enable/disable driver
	command invflp,at_inv		;console invert mode command
	command mechst,at_mch		;console mechanical error command
	command stlste,at_ste		;console stall limit command, el
	command stlsta,at_sta		;console stall limit command, az

	call at_null			;all other commands, no-op
;
;	command error has occured
;

at_ret:	pop	ds
	iret

at_svc	endp

gt_svc	proc				;Graftrak conversion to KCT
	push	ds
	push	cs
	pop	ds			;set up to local ds
	cmp	al,2			;read only function?
	je	gt_rd
	cmp	al,0			;write function?
	je	gt_wr
	jmp	gt_rt			;all others, no-op

gt_wr:	cmp	ch,1			;observer 1?
	je	gt_wr0
	cmp	ch,0			;or none
	je	gt_wr0
	jmp	gt_rt			;if neither, no-op
gt_wr0:	mov	cx,dx			;mov elevation to KCT location
	mov	dx,bx			;mov azimuth to KCT location
	mov	ah,4			;KCT write command
	int	int_at			;now drive the KCT interface
					;and proceed to read function
gt_rd:	mov	ah,3			;KCT read command
	int	int_at
	mov	bx,dx			;mov azimuth from KCT to GT
	mov	dx,cx			;mov elevation from KCT to GT
	cmp	bx,ATNULL
	jne	gt_rd0
	mov	bx,GTNULL		;change clear to GT standard
gt_rd0:	cmp	dx,ATNULL
	jne	gt_rt
	mov	dx,GTNULL		;change clear to GT standard

gt_rt:	pop	ds
	iret				;outta here
gt_svc	endp

at_stat proc	near			;status report
	call	at_null			;clear everything
	mov	bh,01			;as specified
	ret
at_stat	endp

at_read	proc	near
	mov	ax,oldel		;report what we think we've most
	mov	bx,oldaz		;recently done
	mov	cx,ax
	mov	dx,bx
	ret
at_read	endp

at_writ	proc	near
	cmp	oldel,cx
	jne	atw0
	cmp	oldaz,dx
	jne	atw0
	jmp	short	at_read		;if nothing has changed
atw0:	push	cx			;elevation
	mov	oldaz,dx		;new azimuth for report
	mov	oldel,cx		;new elevation for report
	mov	ax,dx			;calculate number for azimuth command
	add	ax,azcorr		;apply azimuth correction
	jge	atw3			;if the answer was negative
	add	ax,360d			;resolve the circle
atw3:	cmp	ax,360d
	jle	atw4			;if this goes greater than 360
	sub	ax,360d			;resolve the circle backwards
atw4:	cmp	onback,00h		;inverted track?
	jne	atw1
	sub	ax,180d			;non inverted case
	jg	atw1			; map 0-180 to 1.8-3.6 v.
	add	ax,360d			;   180-360 to 0.0-1.8 v.
atw1:	cmp	ax,azlostl		;check limits
	jnl	atwal
	mov	ax,azlostl		;limit exceeded, go only to limit
atwal:	cmp	ax,azhistl
	jng	atwah
	mov	ax,azhistl		;limit exceeded, go only to limit
atwah:	mov	dl,3d			;inverted case map 0-360 to 0.0-3.6 v.
					; effect is having moved half circle
	div	dl			;scale to 0-120 range
					; (dx destroyed anyway)
	xor	ah,ah			;clear the remainder
	cmp	ax,azcmd		;a different azimuth can result in the
	je	atw5			; same command, if so, dont perform
	mov	azcmd,ax		;store the new command
	strobe				;write the azimuth
atw5:	pop	ax			;recover elevation value
	cmp	onback,00h
	je	atw10
	mov	dx,ax
	mov	ax,180d			;the inverted case, flip elevation
	sub	ax,dx
atw10:	add	ax,elcorr		;apply mechanical correction
	cmp	ax,ellostl		;check limits
	jnl	atwel
	mov	ax,ellostl		;limit exceeded, go only to limit
atwel:	cmp	ax,elhistl
	jng	atweh			;(note:  jng == jle)
	mov	ax,elhistl		;limit exceeded, go only to limit
atweh:	sal	ax,1 			;now process value, elevation value * 2
	mov	dl,3d
	div	dl			; / 3, now elevation in the range 0-120
	xor	ah,ah			;clear remainder
	or	al,80h			;flag for elevation control
	cmp	ax,elcmd
	jne	atwd
	jmp	at_read			;no need to command a move
atwd:	mov	elcmd,ax
	strobe				;write the elevation
	jmp	at_read			;and report back
at_writ	endp

at_atbl	proc	near
	call	at_null
	mov	al,2			;as desired
	ret
at_atbl	endp

at_null	proc	near			;does nothing, clears registers
	xor	ax,ax
	mov	bx,ax
	mov	cx,ax
	mov	dx,ax
	ret
at_null	endp

at_inv	proc	near			;toggle inverted mode
	xor	onback,0FFh		;flip
	ret
at_inv	endp

at_mch	proc	near
	mov	elcorr,cx 		;store the mechanical correction values
	mov	azcorr,dx		;KCT convention
	mov	oldaz,ATNULL		;Clear current settings since they are
	mov	oldel,ATNULL		; invalidated by new corrections or
	ret
at_mch	endp

at_ste	proc	near
	mov	ellostl,dx		;store mechanical stall limits
	mov	elhistl,cx
	mov	oldel,ATNULL		; invalidated by new corrections or
	ret
at_ste	endp

at_sta	proc	near
	mov	azlostl,dx		;KCT convention
	sub	azlostl,180d		;checked this way because of wrap point
	mov	azhistl,cx
	add	azhistl,180d		;checked this way because of wrap point
	mov	oldaz,ATNULL		;Clear current setting
	ret
at_sta	endp


at_end	label	byte			;sets tsr memory save location

code	ends

	end	start			;the routine to run when run
